from abc import ABC, abstractmethod
from baypy.model import Model


class Regression(ABC):
    """:py:class:`Regression <baypy.regression.regression.Regression>`
    object. \n
    Abstract base class for creating regression objects.

    .. admonition:: See Also
       :class: seealso

       :py:class:`LinearRegression <baypy.regression.linear_regression.LinearRegression>`
    """

    @staticmethod
    @abstractmethod
    def sample(
        model: Model,
        n_iterations: int,
        burn_in_iterations: int,
        n_chains: int,
        seed: int = None
    ) -> None:
        if not isinstance(model, Model):
            raise TypeError(
                f"Parameter 'model' must be an instance of "
                f"'{Model.__module__}.{Model.__name__}'"
            )

        if model.data is None:
            raise ValueError("Missing 'data' in 'model'")

        if model.response_variable is None:
            raise ValueError("Missing 'response_variable' in 'model'")

        if model.response_variable not in model.data.columns:
            raise ValueError(
                f"Column '{model.response_variable}' not found in 'data'"
            )

        if model.priors is None:
            raise ValueError("Missing 'priors' in 'model'")

        for prior in model.priors.keys():
            if (prior not in ['intercept', 'variance']) and \
               (prior not in model.data.columns):
                raise ValueError(f"Column '{prior}' not found in 'model.data'")

        if not isinstance(n_iterations, int):
            raise TypeError("Parameter 'n_iteration' must be an integer")

        if not isinstance(burn_in_iterations, int):
            raise TypeError(
                "Parameter 'burn_in_iterations' must be an integer"
            )

        if not isinstance(n_chains, int):
            raise TypeError("Parameter 'n_chains' must be an integer")

        if (not isinstance(seed, int)) and (seed is not None):
            raise TypeError("Parameter 'seed' must be an integer")

        if n_iterations <= 0:
            raise ValueError("Parameter 'n_iteration' must be greater than 0")

        if burn_in_iterations < 0:
            raise ValueError(
                "Parameter 'burn_in_iterations' must be greater than or equal "
                "to 0"
            )

        if n_chains <= 0:
            raise ValueError("Parameter 'n_chains' must be greater than 0")

        if seed is not None:
            if (seed < 0) or (seed > 2**32 - 1):
                raise ValueError(
                    "Parameter 'seed' must be between 0 and 2**32 - 1"
                )



import numpy as np
import pandas as pd
from ..utils import flatten_matrix


def summary(
    posteriors: dict[str, np.ndarray],
    alpha: float = 0.05,
    quantiles: list[float] = None,
    print_summary: bool = True
) -> dict[str, int | str]:
    """It prints a statistical summary for each posterior.

    Parameters
    ----------
    ``posteriors`` : :py:class:`dict`
        Posterior samples. Posteriors and relative samples are key-value pairs.
        Each sample is a :py:class:`numpy.ndarray` with a number of rows equal
        to the number of iterations and a number of columns equal to the number
        of Markov chains.
    ``alpha`` : :py:class:`float`, optional
        Significance level. It is used to compute the Highest Posterior Density
        (HPD) interval. It must be between ``0`` and ``1``. Default is
        ``0.05``.
    ``quantiles`` : :py:class:`list`, optional
        List of the quantiles to compute, for each posterior. It cannot be
        empty. It must contain only float between ``0`` and ``1``. Default is
        ``[0.025, 0.25, 0.5, 0.75, 0.975]``.
    ``print_summary`` : :py:class:`bool`, optional
        If ``True`` prints the statistical posterior summary report. Default is
        ``True``.

    Returns
    -------
    :py:class:`dict`
        Dictionary with statistical summary of posteriors. It contains:
            - key ``'n_chain'``, the number of Markov chains,
            - key ``'n_iterations'``, the number of regression iterations,
            - key ``'summary'``, the statistical summary of the posteriors, as
              a :py:class:`pandas.DataFrame`,
            - key ``'quantiles'``, quantiles summary of the posteriors, as a
              :py:class:`pandas.DataFrame`.

    .. admonition:: Raises
       :class: warning

       ``TypeError``
           - If ``posteriors`` is not a :py:class:`dict`,
           - if a posterior sample is not a :py:class:`numpy.ndarray`,
           - if ``alpha`` is not a :py:class:`float`,
           - if ``quantiles`` is not a :py:class:`list`,
           - if a ``quantiles`` value is not a :py:class:`float`,
           - if ``print_summary`` is not a :py:class:`bool`.
       ``KeyError``
           If ``posteriors`` does not contain ``'intercept'`` key.
       ``ValueError``
           - If a posterior sample is an empty :py:class:`numpy.ndarray`,
           - if ``alpha`` is not between ``0`` and ``1``,
           - if ``quantiles`` is an empty :py:class:`list`,
           - if a ``quantiles`` value is not between ``0`` and ``1``.

    .. admonition:: See Also
       :class: seealso

       :py:class:`LinearRegression <baypy.regression.linear_regression.LinearRegression>`
    """
    if not isinstance(posteriors, dict):
        raise TypeError("Parameter 'posteriors' must be a dictionary")

    if not all(
        [
            isinstance(posterior_sample, np.ndarray)
            for posterior_sample in posteriors.values()
        ]
    ):
        raise TypeError(
            "All posteriors data must be an instance of 'numpy.ndarray'"
        )

    if not isinstance(print_summary, bool):
        raise TypeError("Parameter 'print_summary' must be a boolean")

    if 'intercept' not in posteriors.keys():
        raise KeyError("Parameter 'posteriors' must contain a 'intercept' key")

    for posterior, posterior_samples in posteriors.items():
        if posterior_samples.size == 0:
            raise ValueError(f"Posterior '{posterior}' data is empty")

    if not isinstance(alpha, float):
        raise TypeError("Parameter 'alpha' must be a float")
    if (alpha < 0) or (alpha > 1):
        raise ValueError("Parameter 'alpha' must be between 0 and 1")

    if quantiles is not None:
        if not isinstance(quantiles, list):
            raise TypeError("Parameter 'quantiles' must be a list")
        if not quantiles:
            raise ValueError("Parameter 'quantiles' cannot be an empty list")
        if not all([isinstance(quantile, float) for quantile in quantiles]):
            raise TypeError("Parameter 'quantiles' must contain only float")
        if any([(quantile < 0) or (quantile > 1) for quantile in quantiles]):
            raise ValueError(
                "Parameter 'quantiles' cannot contain only floats between 0 "
                "and 1"
            )

    quantiles = [0.025, 0.25, 0.5, 0.75, 0.975] if quantiles is None \
        else quantiles

    n_iterations, n_chains = posteriors['intercept'].shape

    general_summary = pd.DataFrame(index=list(posteriors.keys()))
    quantiles_summary = pd.DataFrame(index=list(posteriors.keys()))
    general_summary['Mean'] = np.nan
    general_summary['SD'] = np.nan
    general_summary['HPD min'] = np.nan
    general_summary['HPD max'] = np.nan
    for q in quantiles:
        quantiles_summary[f'{100*q}%'.replace('.0%', '%')] = np.nan

    for variable in general_summary.index:
        general_summary.loc[variable, 'Mean'] = posteriors[variable].mean()
        general_summary.loc[variable, 'SD'] = posteriors[variable].std()
        hpdi_min, hpdi_max = _compute_hpd_interval(
            x=np.sort(flatten_matrix(posteriors[variable])),
            alpha=alpha
        )
        general_summary.loc[variable, 'HPD min'] = hpdi_min
        general_summary.loc[variable, 'HPD max'] = hpdi_max
        for q in quantiles:
            quantiles_summary.loc[
                variable,
                f'{100*q}%'.replace('.0%', '%')
            ] = np.quantile(
                flatten_matrix(posteriors[variable]),
                q
            )

    credibility_mass = f'{100*(1 - alpha)}%'.replace('.0%', '%')

    if print_summary:
        summary_output = (
            f"Number of chains:      {n_chains:>6}\n"
            f"Sample size per chian: {n_iterations:>6}\n"
            f"\n"
            f"Empirical mean, standard deviation, {credibility_mass} HPD "
            f"interval for each variable:\n"
            f"\n"
            f"{general_summary.to_string()}\n"
            f"\n"
            f"Quantiles for each variable:\n"
            f"\n"
            f"{quantiles_summary.to_string()}"
        )
        print(summary_output)

    return {
        'n_chains': n_chains,
        'n_iterations': n_iterations,
        'summary': general_summary,
        'quantiles': quantiles_summary
    }


def _compute_hpd_interval(x: np.ndarray, alpha: float) -> tuple[float, float]:

    n = len(x)
    credibility_mass = 1 - alpha

    interval_idx_included = int(np.floor(credibility_mass*n))
    n_intervals = n - interval_idx_included
    interval_width = x[interval_idx_included:] - x[:n_intervals]
    min_idx = np.argmin(interval_width)
    hpdi_min = x[min_idx]
    hpdi_max = x[min_idx + interval_idx_included]

    return hpdi_min, hpdi_max



import matplotlib.pyplot as plt
from ..model import Model


def residuals_plot(model: Model) -> None:
    r"""It plots the residuals :math:`\epsilon` with respect to predicted
    values :math:`\hat{y}`.

    Parameters
    ----------
    ``model`` : :py:class:`Model <baypy.model.model.Model>`
        The model with data, regressors, response variable and priors to be
        solved through Monte Carlo sampling.

    .. admonition:: Raises
       :class: warning

       ``TypeError``
           If ``model`` is not a :py:class:`Model <baypy.model.model.Model>`.
       ``ValueError``
           - If a ``model.posteriors`` is :py:obj:`None` because the sampling
             has not been done yet,
           - if a posterior key is not a column of ``model.data``,
           - if ``model.data`` is an empty :py:class:`pandas.DataFrame`,
           - if ``model.response_variable`` is not a column of ``model.data``.

    .. admonition:: Notes
       :class: tip

       Predicted values are computed at data points :math:`X` using the
       posteriors means for each regressor's parameter. In the case of linear
       model:

       .. math::
           \hat{y_i} = \beta_0 + \sum_{j = 1}^{m} \beta_j x_{i,j}

       while residuals are the difference between the observed values and the
       predicted values of the ``response_variable``:

       .. math::
           \epsilon_i = y_i - \hat{y_i}

    .. admonition:: See Also
       :class: seealso

       :py:class:`LinearRegression <baypy.regression.linear_regression.LinearRegression>`
    """
    if not isinstance(model, Model):
        raise TypeError(
            f"Parameter 'model' must be an instance of "
            f"'{Model.__module__}.{Model.__name__}'"
        )

    if model.posteriors is None:
        raise ValueError(
            "Posteriors not available, run "
            "'baypy.regression.Regression.sample' to generate posteriors"
        )

    for posterior in model.posteriors.keys():
        if (posterior not in ['intercept', 'variance']) and \
           (posterior not in model.data.columns):
            raise ValueError(f"Column '{posterior}' not found in 'model.data'")

    if model.data.empty:
        raise ValueError(
            "Parameter 'model.data' cannot be an empty 'pandas.DataFrame'"
        )

    if model.response_variable not in model.data.columns:
        raise ValueError(
            f"Column '{model.response_variable}' not found in 'model.data'"
        )

    data = model.residuals()

    _, ax = plt.subplots()

    ax.plot(
        data['predicted'],
        data['residuals'],
        marker='o',
        linestyle='',
        alpha=0.5
    )

    ax.set_xlabel('Predicted')
    ax.set_ylabel('Residuals')

    ax.tick_params(
        bottom=False,
        top=False,
        left=False,
        right=False
    )

    plt.tight_layout()

    plt.show()



from .model import Model
import pandas as pd
import numpy as np
from scipy.stats import norm
from ..utils import flatten_matrix, matrices_to_frame, dot_product


class LinearModel(Model):
    r""":py:class:`LinearModel <baypy.model.linear_model.LinearModel>` object.

    Attributes
    ----------
    :py:attr:`data` : :py:class:`pandas.DataFrame`
        Data for the linear regression model, is a :py:class:`pandas.DataFrame`
        containing all regressor variables :math:`X` and the response variable
        :math:`y`.
    :py:attr:`response_variable` : :py:class:`str`
        Response variable :math:`y` of the linear model.
    :py:attr:`priors` : :py:class:`dict`
        Priors for the regressors' and variance parameters.
    :py:attr:`variable_names` : :py:class:`list`
        The list of all model variables: the regressors :math:`X`, including
        the ``'intercept'`` and the ``'variance'`` :math:`\sigma^2`.
    :py:attr:`posteriors` : :py:class:`dict`
        Posterior samples. Posteriors and relative samples are key-value pairs.
        Each sample is a :py:class:`numpy.ndarray` with a number of rows equals
        to the number of iterations and a number of columns equal to the number
        of Markov chains.

    Methods
    -------
    :py:meth:`posteriors_to_frame`
        It organizes the ``posteriors`` in a :py:class:`pandas.DataFrame`.
    :py:meth:`residuals`
        It computes the residuals :math:`\epsilon` with respect to predicted
        values :math:`\hat{y}`.
    :py:meth:`predict_distribution`
        It predicts a posterior distribution for an unobserved values.
    :py:meth:`likelihood`
        It computes the likelihood of observations :py:attr:`response_variable`
        given a model ``'mean'`` and ``'variance'``.
    :py:meth:`log_likelihood`
        It computes the log likelihood of observations
        :py:attr:`response_variable` given a model ``'mean'`` and
        ``'variance'``.
    """

    def __init__(self):

        self.__data = None
        self.__response_variable = None
        self.__priors = None
        self.__variable_names = None
        self.__posteriors = None

    @property
    def data(self) -> pd.DataFrame:
        r"""Data for the linear regression model, is a
        :py:class:`pandas.DataFrame` containing all regressor variables
        :math:`X` and the response variable :math:`y`.

        Returns
        -------
        :py:class:`pandas.DataFrame`
            Observed data of the model. It cannot be empty. It must contain
            regressor variables :math:`X` and the :py:attr:`response_variable`
            :math:`y`.

        .. admonition:: Raises
           :class: warning

           ``TypeError``
               If :py:attr:`data` is not an instance of
               :py:class:`pandas.DataFrame`.
           ``ValueError``
               If :py:attr:`data` is an empty :py:class:`pandas.DataFrame`.
        """
        return self.__data

    @data.setter
    def data(self, data: pd.DataFrame) -> None:
        super(LinearModel, type(self)).data.fset(self, data)
        self.__data = data

    @property
    def response_variable(self) -> str:
        r"""Response variable :math:`y` of the linear model.

        Returns
        -------
        :py:class:`str`
            Name of the response variable :math:`y`. In must be one of the
            columns of :py:attr:`data`.

        .. admonition:: Raises
           :class: warning

           ``TypeError``
               If :py:attr:`response_variable` is not a :py:class:`str`.
        """
        return self.__response_variable

    @response_variable.setter
    def response_variable(self, response_variable: str) -> None:
        super(
            LinearModel, type(self)
        ).response_variable.fset(self, response_variable)
        self.__response_variable = response_variable

    @property
    def priors(self) -> dict[str, dict[str, float | int]]:
        r"""Priors for the regressors' and variance parameters.
        Each prior is a key-value pair, where the value is a :py:class:`dict`
        with:

        - hyperparameter names as keys
        - hyperparameter values as values.

        Returns
        -------
        :py:class:`dict`
            Priors for each random variable. It must contain an ``'intercept'``
            and a ``'variance'`` keys. Each value must be a :py:class:`dict`
            with hyperparameter names as key and hyperparameter values as
            values.

        .. admonition:: Raises
           :class: warning

           ``TypeError``
               - If :py:attr:`priors` is not a :py:class:`dict`,
               - if a :py:attr:`priors`' value is not a :py:class:`dict`.
           ``ValueError``
               - If :py:attr:`priors` is an empty :py:class:`dict`,
               - if a :py:attr:`priors`' value is an empty :py:class:`dict`,
               - if a ``'variance'`` value is not positive,
               - if a ``'shape'`` value is not positive,
               - if a ``'scale'`` value is not positive.
           ``KeyError``
               - If :py:attr:`priors` does not contain both ``'intercept'``
                 and ``'variance'`` keys,
               - if a prior's hyperparameters are not:
                   + ``'mean'`` and ``'variance'`` for a regression parameter
                     :math:`\beta_j` or
                   + ``'shape'`` and ``'scale'`` for ``variance``
                     :math:`\sigma^2`.

        .. admonition:: Notes
           :class: tip

           To each random variables is assigned a prior distribution:

           - to each regressor parameter :math:`\beta_j` is assigned a normal
             prior distribution with hyperparameters ``'mean'``
             :math:`\beta_j^0` and ``'variance'`` :math:`\Sigma_{\beta_j}^0`:

             .. math::
               \beta_j \sim N(\beta_j^0 , \Sigma_{\beta_j}^0)

           - to variance :math:`\sigma^2` is assigned an inverse gamma
             distribution with hyperparameters ``'shape'`` :math:`\kappa^0`
             and ``'scale'`` :math:`\theta^0`:

             .. math::
               \sigma^2 \sim \text{Inv-}\Gamma(\kappa^0, \theta^0)

        .. admonition:: Examples
           :class: important

           Consider a linear regression of the :py:attr:`response_variable`
           :math:`y` with respect to regressors :math:`x_1`, :math:`x_2` and
           :math:`x_3`, according to the following model:

           .. math::
               y \sim N(\mu, \sigma^2)
           .. math::
               \mu = \beta_0 + \beta_1 x_1 + \beta_2 x_2 + \beta_3 x_3

           then the sampler would require priors for:
               - parameter :math:`\beta_0` of variable ``'intercept'``, with
                 ``'mean'`` :math:`\beta_0^0` and ``'variance'``
                 :math:`\Sigma_{\beta_0}^0`
               - parameter :math:`\beta_1` of variable :math:`x_1`, with
                 ``'mean'`` :math:`\beta_1^0` and ``'variance'``
                 :math:`\Sigma_{\beta_1}^0`
               - parameter :math:`\beta_2` of variable :math:`x_2`, with
                 ``'mean'`` :math:`\beta_2^0` and ``'variance'``
                 :math:`\Sigma_{\beta_2}^0`
               - parameter :math:`\beta_3` of variable :math:`x_3`, with
                 ``'mean'`` :math:`\beta_3^0` and ``'variance'``
                 :math:`\Sigma_{\beta_3}^0`
               - variable :math:`\sigma^2`, with ``'shape'`` :math:`\kappa^0`
                 and ``'scale'`` :math:`\theta^0`

           >>> model = baypy.model.LinearModel()
           >>> model.priors = {
           ...     'intercept': {'mean': 0, 'variance': 1e6},
           ...     'x_1': {'mean': 0, 'variance': 1e6},
           ...     'x_2': {'mean': 0, 'variance': 1e6},
           ...     'x_3': {'mean': 0, 'variance': 1e6},
           ...     'variance': {'shape': 1, 'scale': 1e-6}
           ... }
        """
        return self.__priors

    @priors.setter
    def priors(self, priors: dict[str, dict[str, float | int]]) -> None:
        super(LinearModel, type(self)).priors.fset(self, priors)

        if 'variance' not in priors.keys():
            raise KeyError("Parameter 'priors' must contain a 'variance' key")

        for prior, values in priors.items():
            if not isinstance(values, dict):
                raise TypeError(
                    f"The value of prior '{prior}' must be a dictionary"
                )
            if len(values) == 0:
                raise ValueError(
                    f"The value of prior '{prior}' cannot be an empty "
                    f"dictionary"
                )
            if prior != 'variance':
                if set(values.keys()) != {'mean', 'variance'}:
                    raise KeyError(
                        f"The value of prior '{prior}' must be a dictionary "
                        f"containing 'mean' and 'variance' keys only"
                    )
                if values['variance'] <= 0:
                    raise ValueError(
                        f"The 'variance' of prior '{prior}' must be positive"
                    )
            else:
                if set(values.keys()) != {'shape', 'scale'}:
                    raise KeyError(
                        f"The value of prior '{prior}' must be a dictionary "
                        f"containing 'shape' and 'scale' keys only"
                    )
                for parameter in ['shape', 'scale']:
                    if values[parameter] <= 0:
                        raise ValueError(
                            f"The '{parameter}' of prior '{prior}' must be "
                            f"positive"
                        )

        self.__priors = priors
        self.__variable_names = list(priors.keys())
        self.__variable_names.insert(
            0,
            self.__variable_names.pop(self.__variable_names.index('intercept'))
        )

    @property
    def variable_names(self) -> list[str]:
        r"""Variables of the linear model.

        Returns
        -------
        :py:class:`list`
            The list of all model variables: the regressors :math:`X`,
            including the ``'intercept'`` and the ``'variance'``
            :math:`\sigma^2`.
        """
        return self.__variable_names

    @property
    def posteriors(self) -> dict[str, np.ndarray]:
        r"""Posteriors of the regressors' and variance parameters.
        Posteriors and relative samples are key-value pairs. Each sample is a
        :py:class:`numpy.ndarray` with a number of rows equals to the number of
        iterations and a number of columns equal to the number of Markov
        chains.

        Returns
        -------
        :py:class:`dict`
            Posterior samples. Posteriors and relative samples are key-value
            pairs. Each sample is a :py:class:`numpy.ndarray` with a number of
            rows equals to the number of iterations and a number of columns
            equal to the number of Markov chains.

        .. admonition:: Raises
           :class: warning

           ``TypeError``
               - If :py:attr:`posteriors` is not a :py:class:`dict`,
               - if a posterior sample is not a :py:class:`numpy.ndarray`.
           ``KeyError``
               If :py:attr:`posteriors` does not contain both ``'intercept'``
               and ``'variance'`` keys.
           ``ValueError``
               If a posterior sample is an empty :py:class:`numpy.ndarray`.
        """
        return self.__posteriors

    @posteriors.setter
    def posteriors(self, posteriors: dict[str, np.ndarray]) -> None:
        super(LinearModel, type(self)).posteriors.fset(self, posteriors)

        if 'variance' not in posteriors.keys():
            raise KeyError(
                "Parameter 'posteriors' must contain a 'variance' key"
            )

        self.__posteriors = posteriors

    def posteriors_to_frame(self) -> pd.DataFrame:
        """It organizes the :py:attr:`posteriors` in a
        :py:class:`pandas.DataFrame`. Each posterior is a frame column.
        The length of the frame is the number of sampling iterations times the
        number of sampling chains.

        Returns
        -------
        :py:class:`pandas.DataFrame`
            Returns posterior samples. Posteriors are organized in a
            :py:class:`pandas.DataFrame`, one for each column. The length of
            the frame is the number of sampling iterations times the number of
            sampling chains.

        .. admonition:: Raises
           :class: warning

           ``ValueError``
               If :py:attr:`posteriors` are not available because the method
               :py:meth:`LinearRegression.sample <baypy.regression.linear_regression.LinearRegression.sample>`
               has not been called yet.
        """
        if self.__posteriors is None:
            raise ValueError(
                "Posteriors not available, run "
                "'baypy.regression.LinearRegression.sample' to generate "
                "posteriors"
            )
        return matrices_to_frame(matrices_dict=self.__posteriors)

    def residuals(self) -> pd.DataFrame:
        r"""It computes the residuals :math:`\epsilon` with respect to
        predicted values :math:`\hat{y}`.

        Returns
        -------
        :py:class:`pandas.DataFrame`
            Returns a copy of :py:attr:`data` with 3 more columns:
            ``'intercept'``, ``'predicted'`` and ``'residuals'``.

        .. admonition:: Raises
           :class: warning

           ``ValueError``
               - If :py:attr:`data` is :py:obj:`None` because the property
                 :py:attr:`data` has not been set,
               - if :py:attr:`response_variable` is not a column of
                 :py:attr:`data`,
               - If a :py:attr:`posteriors` is :py:obj:`None` because the
                 sampling has not been done yet.

        .. admonition:: Notes
           :class: tip

           Predicted values are computed at data points :math:`X` using the
           posteriors means for each regressor's parameter:

           .. math::
               \hat{y_i} = \beta_0 + \sum_{j = 1}^{m} \beta_j x_{i,j}

           while residuals are the difference between the observed values and
           the predicted values of the :py:attr:`response_variable`:

           .. math::
               \epsilon_i = y_i - \hat{y_i}
        """
        if self.__data is None:
            raise ValueError(
                "Data not available, set data with "
                "'baypy.model.LinearModel.data"
            )

        if self.__response_variable not in self.__data.columns:
            raise ValueError(
                f"Column '{self.__response_variable}' not found in 'data'"
            )

        if self.__posteriors is None:
            raise ValueError(
                "Posteriors not available, run "
                "'baypy.regression.LinearRegression.sample' to generate "
                "posteriors"
            )

        posterior_means = {
            posterior: flatten_matrix(posterior_samples).mean()
            for posterior, posterior_samples in self.__posteriors.items()
            if posterior != 'variance'
        }

        data = self.__data.copy()
        data['intercept'] = 1
        data['predicted'] = dot_product(data=data, regressors=posterior_means)
        data['residuals'] = data[self.__response_variable] - data['predicted']

        return data

    def predict_distribution(
        self,
        predictors: dict[str, float | int]
    ) -> np.ndarray:
        """It predicts a posterior distribution for an unobserved values. For
        each posterior sample, it draws a sample from the likelihood.

        Parameters
        ----------
        ``predictors`` : :py:class:`dict`
            Values of predictors :math:`X` at which compute the posterior
            distribution. Each predictor has to be set as a key-value pair.

        Returns
        -------
        :py:class:`numpy.ndarray`
            Array of the predicted posterior distribution. It contains a
            number of element equal to the number of regression iterations
            times the number of model Markov chains.

        .. admonition:: Raises
           :class: warning

           ``TypeError``
               If ``predictors`` is not a :py:class:`dict`.
           ``KeyError``
               If a ``predictors`` key is not a key of :py:attr:`posteriors`.
           ``ValueError``
               If ``predictors`` is an empty :py:class:`dict`.

        .. admonition:: See Also
           :class: seealso

           :py:class:`LinearRegression <baypy.regression.linear_regression.LinearRegression>`
        """
        super().predict_distribution(predictors=predictors)

        for regressor in predictors.keys():
            if regressor not in self.__posteriors.keys():
                raise KeyError(
                    f"Regressor '{regressor}' not found in 'posteriors' keys"
                )

        prediction = matrices_to_frame(matrices_dict=self.__posteriors)
        prediction['mean'] = prediction['intercept'] + dot_product(
            data=prediction,
            regressors=predictors
        )

        return norm.rvs(
            loc=prediction['mean'],
            scale=np.sqrt(prediction['variance']),
            size=len(prediction)
        )

    def likelihood(self, data: pd.DataFrame) -> np.ndarray:
        r"""It computes the likelihood of observations
        :py:attr:`response_variable` given a model ``'mean'`` and
        ``'variance'``.

        Parameters
        ----------
        ``data``: :py:class:`pandas.DataFrame`
            Data to use for likelihood computation. It cannot be empty. It must
            contain columns :py:attr:`response_variable`, ``'mean'`` and
            ``'variance'``.

        Returns
        -------
        :py:class:`numpy.ndarray`
            Array of computed likelihood. It has the same length of ``data``.
            Each element is a likelihood computation of each row of ``data``.

        .. admonition:: Raises
           :class: warning

           ``TypeError``
               If ``data`` is not an instance of :py:class:`pandas.DataFrame`.
           ``ValueError``
               - If ``data`` is an empty :py:class:`pandas.DataFrame`,
               - if :py:attr:`response_variable` is not a column of ``data``,
               - if ``'mean'`` is not a column of ``data``,
               - if ``'variance'`` is not a column of ``data``.

        .. admonition:: Notes
           :class: tip

           The likelihood is computed with the normal distribution probability
           density function:

           .. math::
               L(y) = \frac{1}{\sqrt{2 \pi \sigma^2}} \exp{- \frac{\left(y -
               \mu \right)^2}{2 \sigma^2}}

           where :math:`\mu` is the ``'mean'`` column and :math:`\sigma^2` is
           the ``'variance'`` column.
        """
        super().likelihood(data=data)

        for col in [self.__response_variable, 'mean', 'variance']:
            if col not in data.columns:
                raise ValueError(f"Column '{col}' not found in 'data'")

        return norm.pdf(
            x=data[self.__response_variable],
            loc=data['mean'],
            scale=np.sqrt(data['variance'])
        )

    def log_likelihood(self, data: pd.DataFrame) -> np.ndarray:
        r"""It computes the log likelihood of observations
        :py:attr:`response_variable` given a model ``'mean'`` and
        ``'variance'``.

        Parameters
        ----------
        ``data``: :py:class:`pandas.DataFrame`
            Data to use for log likelihood computation. It cannot be empty. It
            must contain columns :py:attr:`response_variable`, ``'mean'`` and
            ``'variance'``.

        Returns
        -------
        :py:class:`numpy.ndarray`
            Array of computed log likelihood. It has the same length of
            ``data``. Each element is a log likelihood computation of each row
            of ``data``.

        .. admonition:: Raises
           :class: warning

           ``TypeError``
               If ``data`` is not an instance of :py:class:`pandas.DataFrame`.
           ``ValueError``
               - If ``data`` is an empty :py:class:`pandas.DataFrame`,
               - if :py:attr:`response_variable` is not a column of ``data``,
               - if ``'mean'`` is not a column of ``data``,
               - if ``'variance'`` is not a column of ``data``.

        .. admonition:: Notes
           :class: tip

           The log likelihood is computed as the log of the normal distribution
           probability density function:

           .. math::
               l(y) = - \frac{1}{2} \log{2 \pi \sigma^2} - \frac{1}{2}
               \frac{\left(y - \mu \right)^2}{\sigma^2}

           where :math:`\mu` is the ``'mean'`` column and :math:`\sigma^2` is
           the ``'variance'`` column.
        """
        super().log_likelihood(data=data)

        for col in [self.__response_variable, 'mean', 'variance']:
            if col not in data.columns:
                raise ValueError(f"Column '{col}' not found in 'data'")

        return norm.logpdf(
            x=data[self.__response_variable],
            loc=data['mean'],
            scale=np.sqrt(data['variance'])
        )

    def _compute_model_parameters_at_posterior_means(self) -> pd.DataFrame:

        posterior_means = {
            posterior: flatten_matrix(posterior_samples).mean()
            for posterior, posterior_samples in self.__posteriors.items()
            if posterior != 'variance'
        }

        data = self.__data.copy()
        data['intercept'] = 1
        data['mean'] = dot_product(data=data, regressors=posterior_means)
        data['variance'] = flatten_matrix(self.__posteriors['variance']).mean()

        return data

    def _compute_model_parameters_at_observation(self, i: int) -> pd.DataFrame:

        posterior_means = {
            posterior: posterior_samples[i, :].mean()
            for posterior, posterior_samples in self.__posteriors.items()
            if posterior != 'variance'
        }

        data = self.__data.copy()
        data['intercept'] = 1
        data['mean'] = dot_product(data=data, regressors=posterior_means)
        data['variance'] = self.__posteriors['variance'][i, :].mean()

        return data



import numpy as np
import pandas as pd


def flatten_matrix(matrix: np.ndarray) -> np.ndarray:
    """Flattens a matrix of dimensions ``(M, N)`` to ``(M*N, )``.

    Parameters
    ----------
    matrix : numpy.ndarray
        Matrix with ``M`` rows and ``N`` columns.

    Returns
    -------
    numpy.ndarray
        Flatten array with ``M*N`` elements.

    .. admonition:: Raises
       :class: warning

       TypeError
           If ``matrix`` is not a ``numpy.ndarray``.
       ValueError
           If ``matrix`` is an empty ``numpy.ndarray``.

    .. admonition:: Examples
       :class: important

       >>> import numpy
       >>> a = numpy.array([[1, 2], [3, 4], [5, 6]])
       >>> a
       array([[1, 2],
              [3, 4],
              [5, 6]])
       >>> flatten_matrix(a)
       array([1, 2, 3, 4, 5, 6])
    """
    if not isinstance(matrix, np.ndarray):
        raise TypeError(
            "Parameter 'matrix' must be an instance of 'numpy.ndarray'"
        )

    if matrix.size == 0:
        raise ValueError("Parameter 'matrix' is empty")

    return np.asarray(matrix).reshape(-1)


def matrices_to_frame(matrices_dict: dict[str, np.ndarray]) -> pd.DataFrame:
    """Organizes a dictionary of matrices in a ``pandas.DataFrame``. Each
    matrix becomes a frame column, with column name equal to the matrix'
    relative key in the dictionary. If the matrix has dimensions ``(M, N``),
    then the relative frame column has length ``M*N``.

    Parameters
    ----------
    matrices_dict : dict
        Dictionary of matrices to be organized. Matrices and relative names are
        key-value pairs. Each matrix is a ``numpy.ndarray`` with dimensions
        ``(M, N)``.

    Returns
    -------
    pandas.DataFrame
        Reorganized matrices frame. Matrices are organized in a
        ``pandas.DataFrame``, one for each column. The length of the frame is
        ``M*N``.

    .. admonition:: Raises
       :class: warning

       TypeError
           - If ``matrices_dict`` is not a ``dict``,
           - if a ``matrices_dict`` value is not a ``numpy.ndarray``.
       ValueError
           If a ``matrices_dict`` value is an empty ``numpy.ndarray``.

    .. admonition:: Examples
       :class: important

       >>> import numpy
       >>> a = numpy.array([[1, 2], [3, 4], [5, 6]])
       >>> b = numpy.array([[7, 8], [9, 10], [11, 12]])
       >>> d = {'a': a, 'b': b}
       >>> matrices_to_frame(d)
          a   b
       0  1   7
       1  2   8
       2  3   9
       3  4  10
       4  5  11
       5  6  12
    """
    if not isinstance(matrices_dict, dict):
        raise TypeError("Parameter 'matrices_dict' must be a dictionary")

    for matrix_name, matrix in matrices_dict.items():
        if not isinstance(matrix, np.ndarray):
            raise TypeError(
                f"Matrix '{matrix_name}' must be an instance of "
                f"'numpy.ndarray'"
            )

        if matrix.size == 0:
            raise ValueError(f"Matrix '{matrix_name}' is empty")

    return pd.DataFrame(
        {col: flatten_matrix(matrix) for col, matrix in matrices_dict.items()}
    )


def dot_product(
    data: pd.DataFrame,
    regressors: dict[str, float | int]
) -> np.ndarray:
    """Computes the dot product between columns of ``data`` and values of
    ``regressors``.

    Parameters
    ----------
    data : pandas.DataFrame
        The dataframe to be used for the dot product. It cannot be empty. It
        must contain all ``regressors`` keys.

    regressors : dict
        Dictionary with regressors' values. It cannot be empty. All regressors
        and relative values must be a key-value pair.

    Returns
    -------
    numpy.ndarray
        Array of computed dot product. Each element is the dot product of a
        ``data`` row with respect to each ``regressors``. It has the same
        length of ``data``.

    .. admonition:: Raises
       :class: warning

       TypeError
           - If ``data`` is not a ``pandas.DataFrame``,
           - if ``regressors`` is not a ``dict``,
           - if a ``regressors`` value is not a ``int`` or a ``float``.
       KeyError
           If a ``regressors`` key is not a column of ``data``.
       ValueError
           - If ``data`` is an empty ``pandas.DataFrame``,
           - if ``regressors`` is an empty ``dict``.

    .. admonition:: Examples
       :class: important

       >>> import pandas as pd
       >>> data = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
       >>> regressors = {'a': 2, 'b': -1}
       >>> dot_product(data = data, regressors = regressors)
       array([-2, -1,  0])
    """
    if not isinstance(data, pd.DataFrame):
        raise TypeError(
            "Parameter 'data' must be an instance of 'pandas.DataFrame'"
        )

    if data.empty:
        raise ValueError(
            "Parameter 'data' cannot be an empty 'pandas.DataFrame'"
        )

    if not isinstance(regressors, dict):
        raise TypeError("Parameter 'regressors' must be a dictionary")

    if len(regressors) == 0:
        raise ValueError(
            "Parameter 'regressors' cannot be an empty dictionary"
        )

    for regressor in regressors.keys():
        if regressor not in data.columns:
            raise KeyError(f"Column '{regressor}' not found in 'data'")

    if not all(
        [
            isinstance(regressor, float | int)
            for regressor in regressors.values()
        ]
    ):
        raise TypeError("All 'regressors' values must be integers or float.")

    data = data[regressors.keys()]
    return np.dot(data, list(regressors.values()))



from ..model import Model
import numpy as np


def compute_DIC(model: Model, print_summary: bool = True) -> dict[str, float]:
    r"""It computes and prints the Deviance Information Criterion (DIC) for
    the fitted model.

    Parameters
    ----------
    ``model`` : :py:class:`Model <baypy.model.model.Model>`
        The model with data, regressors, response variable and priors to be
        solved through Monte Carlo sampling.
    ``print_summary`` : :py:class:`bool`, optional
        If ``True`` prints the deviance summary report. Default is ``True``.

    Returns
    -------
    :py:class:`dict`
        Dictionary with deviance summary. It contains:
            - key ``'Deviance at posterior means'``,
            - key ``'Posterior mean deviance'``,
            - key ``'Effective number of parameters'``,
            - key ``'Deviance Information Criterion'``.

    .. admonition:: Raises
       :class: warning

       ``TypeError``
           - If ``model`` is not a :py:class:`Model <baypy.model.model.Model>`,
           - if ``print_summary`` is not a :py:class:`bool`.
       ``ValueError``
           - If a ``model.posteriors`` is :py:obj:`None` because the sampling
             has not been done yet,
           - if a posterior key is not a column of ``model.data``,
           - if ``model.data`` is an empty :py:class:`pandas.DataFrame`,
           - if ``model.response_variable`` is not a column of ``model.data``.

    .. admonition:: Notes
       :class: tip

       The DIC measures posterior predictive error by penalizing the fit of a
       model (deviance) by its complexity, determined by the effective number
       of parameters.
       Comparing some alternative models, the smaller the DIC of a model, the
       *better* the model.
       Consider a linear regression of the response variable :math:`y` with
       respect to regressors :math:`X`, according to the following model:

       .. math::
           y \sim N(\mu, \sigma^2)
       .. math::
           \mu = \beta_0 + B X = \beta_0 + \sum_{j = 1}^m \beta_j x_j

       then the *likelyhood* is:

       .. math::
           p \left( y \left\vert B,\sigma^2 \right. \right) = \frac{1}{\sqrt{2
           \pi \sigma^2}} \exp{- \frac{\left(y - \mu \right)^2}{2 \sigma^2}} .

       The *deviance* [1]_ [2]_ is defined as:

       .. math::
           D \left( y, B, \sigma^2 \right) = -2\log p \left( y \left\vert B,
           \sigma^2 \right. \right) .

       The *deviance* at posterior mean of :math:`B` and :math:`\sigma^2`,
       denoted by :math:`\overline{B}` and :math:`\overline{\sigma^2}` is:

       .. math::
           D_{{\overline{\beta}}, \overline{\sigma^2}} (y) = D \left( y,
           \overline{B}, \overline{\sigma^2} \right)

       while the posterior mean deviance is:

       .. math::
           \overline{D} \left( y, B, \sigma^2 \right) = E \left( D(y, B,
           \sigma^2) \left. \right\vert y \right) .

       and the *effective number of parameter* is defined as:

       .. math::
           pD = \overline{D} \left( y, B, \sigma^2 \right) -
           D_{{\overline{\beta}}, \overline{\sigma^2}} (y) .

       The *Deviance Information Criterion* [1]_ is:

       .. math::
           DIC = 2 \overline{D} \left( y, B, \sigma^2 \right) -
           D_{{\overline{\beta}}, \overline{\sigma^2}} (y) =
           \overline{D} \left( y, B, \sigma^2 \right) + pD =
           D_{{\overline{B}}, \overline{\sigma^2}} (y) + 2pD .

    .. admonition:: References
       :class: note

       .. [1] O. Spiegelhalter DJ, Best NG, Carlin BP, van der Linde A.
          Bayesian measures of model complexity and fit. J R Statist Soc B.
          2002;64:583–616.
       .. [2] Gelman A, Carlin JB, Stern HS, Rubin DS. Bayesian Data Analysis.
          2. Chapman & Hall/CRC; Boca Raton, Florida: 2004.

    .. admonition:: See Also
       :class: seealso

       :py:class:`LinearRegression <baypy.regression.linear_regression.LinearRegression>`
    """
    if not isinstance(model, Model):
        raise TypeError(
            f"Parameter 'model' must be an instance of "
            f"'{Model.__module__}.{Model.__name__}'"
        )

    if model.posteriors is None:
        raise ValueError(
            "Posteriors not available, run "
            "'baypy.regression.Regression.sample' to generate posteriors"
        )

    if not isinstance(print_summary, bool):
        raise TypeError("Parameter 'print_summary' must be a boolean")

    for posterior, _ in model.posteriors.items():
        if (posterior not in ['intercept', 'variance']) and \
           (posterior not in model.data.columns):
            raise ValueError(f"Column '{posterior}' not found in 'model.data'")

    if model.data.empty:
        raise ValueError(
            "Parameter 'model.data' cannot be an empty 'pandas.DataFrame'"
        )

    if model.response_variable not in model.data.columns:
        raise ValueError(
            f"Column '{model.response_variable}' not found in 'model.data'"
        )

    deviance_at_posterior_means = _compute_deviance_at_posterior_means(
        model=model
    )
    posterior_mean_deviance = _compute_posterior_mean_deviance(model=model)
    effective_number_of_parameters = \
        posterior_mean_deviance - deviance_at_posterior_means
    DIC = effective_number_of_parameters + posterior_mean_deviance

    if print_summary:
        dic_output = (
            f"Deviance at posterior means     "
            f"{deviance_at_posterior_means:>12.2f}\n"
            f"Posterior mean deviance         "
            f"{posterior_mean_deviance:>12.2f}\n"
            f"Effective number of parameters  "
            f"{effective_number_of_parameters:>12.2f}\n"
            f"Deviance Information Criterion  {DIC:>12.2f}"
        )
        print(dic_output)

    return {
        "Deviance at posterior means": deviance_at_posterior_means,
        "Posterior mean deviance": posterior_mean_deviance,
        "Effective number of parameters": effective_number_of_parameters,
        "Deviance Information Criterion": DIC
    }


def _compute_deviance_at_posterior_means(model: Model) -> float:
    data = model._compute_model_parameters_at_posterior_means()
    likelihood = model.likelihood(data=data)

    return -2*np.sum(np.log(likelihood))


def _compute_posterior_mean_deviance(model: Model) -> float:
    deviance = []
    for i in range(model.posteriors['intercept'].shape[0]):
        data = model._compute_model_parameters_at_observation(i)
        likelihood = model.likelihood(data=data)
        deviance.append(-2*np.sum(np.log(likelihood)))

    return np.mean(deviance)



import numpy as np
import pandas as pd
from scipy.stats import invgamma, multivariate_normal
from baypy.utils import flatten_matrix


def sample_sigma2(
    y: pd.Series,
    X: np.ndarray,
    beta: np.ndarray,
    k_1: float,
    theta_0: float
) -> float:
    r"""It draws a sample for the variance from an inverse-gamma conditional
    posterior distribution at each regression iteration.

    Parameters
    ----------
    ``y`` : :py:class:`pandas.Series`
        Vector of the response variable :math:`y`. It has a number of rows
        equal to the number of observed data.
    ``X`` : :py:class:`numpy.ndarray`
        Regressors matrix :math:`X`. It has a number of rows equal to the
        number of observed data and a number of columns equal to the number of
        regressors.
    ``beta`` : :py:class:`numpy.ndarray`
        Regressors' parameters array of the model. It is sampled by the
        function
        :py:func:`sample_beta <baypy.regression.functions.sample_beta>` at each
        iteration.
    ``k_1`` : :py:class:`float`
        Sum of the variance prior ``'shape'`` and the number of observed data.
    ``theta_0`` : :py:class:`float`
        Variance prior ``'scale'``.

    Returns
    -------
    :py:class:`numpy.ndarray`
        Sampled variance :math:`\sigma^2`.

    .. admonition:: Notes
       :class: tip

       The variance is drawn from an inverse-gamma distribution. The
       conditional posterior distribution is:

       .. math::
           H \left( \sigma^2 \left\vert B, y \right. \right) \sim
           \text{Inv-}\Gamma \left( \frac{\kappa_1}{2}, \frac{\theta_1}{2}
           \right)

       where :math:`B` is the :math:`\beta_j` vector, :math:`\kappa_1` is:

       .. math::
           \kappa_1 = \kappa_0 + n

       where :math:`n` is the number of observed data and :math:`\theta_1` is:

       .. math::
           \theta_1 = \theta_0 + \left( y - X B \right)^T \left( y - X B
           \right) .

    .. admonition:: See Also
       :class: seealso

       :py:func:`sample_beta <baypy.regression.functions.sample_beta>`
    """
    y_X_beta = y - flatten_matrix(np.dot(X, beta))
    theta_1 = theta_0 + np.dot(y_X_beta.transpose(), y_X_beta)

    return invgamma.rvs(a=k_1/2, scale=theta_1/2)


def sample_beta(
    Xt_X: np.ndarray,
    Xt_y: np.ndarray,
    sigma2: float,
    Sigma_0_inv: np.ndarray,
    Sigma_0_inv_Beta_0: np.ndarray
) -> np.ndarray:
    r"""it draws a sample for each regressor parameter from a normal
    conditional posterior distribution at each regression iteration.

    Parameters
    ----------
    ``Xt_X`` : :py:class:`numpy.ndarray`
       Dot product of transposed regressors with themselves: :math:`X^T X`. It
       is a symmetric matrix with a number of rows and columns equal to the
       number of regressors.
    ``Xt_y`` : :py:class:`numpy.ndarray`
        Dot product of transposed regressors with the response variable:
        :math:`X^T y`. It is a one-dimensional array with a number of rows
        equal to the number of regressors.
    ``sigma2`` : :py:class:`float`
        Variance of the model. It is sampled by the function
        :py:func:`sample_sigma2 <baypy.regression.functions.sample_sigma2>` at
        each iteration.
    ``Sigma_0_inv`` : :py:class:`numpy.ndarray`
        Regressors' parameters variance priors matrix. It is a symmetric matrix
        with a number of rows and columns equal to the number of regressors. It
        is the inverse of a matrix which has regressors' parameters' prior
        variances on the diagonal and ``0`` everywhere else.
    ``Sigma_0_inv_Beta_0`` : :py:class:`numpy.ndarray`
        Dot product of ``Sigma_0_inv`` with a one-dimensional vector which has
        regressors' parameters' prior means on rows.

    Returns
    -------
    :py:class:`numpy.ndarray`
        Array of the sampled regressors' parameters :math:`B`. It has a number
        of element equal to the number of regressors.

    .. admonition:: Notes
       :class: tip

       The regressors' parameters are drawn from a multivariate normal
       distribution. The conditional posterior distribution is:

       .. math::
           H \left( B \left\vert \sigma^2, y \right. \right) \sim N(M, V)

       where :math:`M` is a one-dimensional vector with a number of rows equal
       to the number of regressors:

       .. math::
           M = \left( \Sigma_0^{-1} + \frac{1}{\sigma^2} X^T X \right)^{-1}
           \left( \Sigma_0^{-1} B_0 + \frac{1}{\sigma^2} X^T y \right)

       and :math:`V` is a symmetric matrix with a number of rows and columns
       equal to the number of regressors:

       .. math::
           V = \left( \Sigma_0^{-1} + \frac{1}{\sigma^2} X^T X \right)^{-1} .

       :math:`B_0` is:

       .. math::
           B_0 =
           \begin{bmatrix}
               \beta_0^0 \\
               \beta_1^0 \\
               \vdots    \\
               \beta_m^0
           \end{bmatrix}

       and :math:`\Sigma_0^{-1}` is:

       .. math::
           \Sigma_0^{-1} =
           \begin{bmatrix}
               \Sigma_{\beta_0}^0 & 0 & \dots & 0 \\
               0 & \Sigma_{\beta_1}^0 & \dots & 0 \\
               \vdots & \vdots & \ddots & 0       \\
               0 & 0 & 0 & \Sigma_{\beta_m}^0
           \end{bmatrix}

    .. admonition:: See Also
       :class: seealso

       :py:func:`sample_sigma2 <baypy.regression.functions.sample_sigma2>`
    """
    V = np.linalg.inv(Sigma_0_inv + Xt_X/sigma2)
    M = np.dot(V, Sigma_0_inv_Beta_0 + Xt_y/sigma2)
    beta = multivariate_normal.rvs(mean=flatten_matrix(M), cov=V, size=1)

    if Xt_X.shape != (1, 1):
        return beta
    else:
        return np.array([beta])



from abc import ABC, abstractmethod
import numpy as np
import pandas as pd


class Model(ABC):
    """:py:class:`Model <baypy.model.model.Model>` object. \n
    Abstract base class for creating model objects.

    .. admonition:: See Also
       :class: seealso

       :py:class:`LinearModel <baypy.model.linear_model.LinearModel>`
    """

    @property
    @abstractmethod
    def data(self) -> None: ...

    @data.setter
    @abstractmethod
    def data(self, data: pd.DataFrame) -> None:
        if not isinstance(data, pd.DataFrame):
            raise TypeError(
                "Parameter 'data' must be an instance of 'pandas.DataFrame'"
            )

        if data.empty:
            raise ValueError(
                "Parameter 'data' cannot be an empty 'pandas.DataFrame'"
            )

    @property
    @abstractmethod
    def response_variable(self) -> None: ...

    @response_variable.setter
    @abstractmethod
    def response_variable(self, response_variable: str) -> None:
        if not isinstance(response_variable, str):
            raise TypeError("Parameter 'response_variable' must be a string")

    @property
    @abstractmethod
    def priors(self) -> None: ...

    @priors.setter
    @abstractmethod
    def priors(self, priors: dict[str, dict[str, float | int]]) -> None:
        if not isinstance(priors, dict):
            raise TypeError("Parameter 'priors' must be a dictionary")

        if len(priors) == 0:
            raise ValueError(
                "Parameter 'priors' cannot be an empty dictionary"
            )

        if 'intercept' not in priors.keys():
            raise KeyError("Parameter 'priors' must contain a 'intercept' key")

    @property
    @abstractmethod
    def variable_names(self) -> None: ...

    @property
    @abstractmethod
    def posteriors(self) -> None: ...

    @posteriors.setter
    @abstractmethod
    def posteriors(self, posteriors: dict[str, np.ndarray]) -> None:
        if not isinstance(posteriors, dict):
            raise TypeError("Parameter 'posteriors' must be a dictionary")

        if not all(
            [
                isinstance(posterior_sample, np.ndarray)
                for posterior_sample in posteriors.values()
            ]
        ):
            raise TypeError(
                "All posteriors data must be an instance of 'numpy.ndarray'"
            )

        if 'intercept' not in posteriors.keys():
            raise KeyError(
                "Parameter 'posteriors' must contain a 'intercept' key"
            )

        for posterior, posterior_samples in posteriors.items():
            if posterior_samples.size == 0:
                raise ValueError(f"Posterior '{posterior}' data is empty")

    @abstractmethod
    def posteriors_to_frame(self) -> None: ...

    @abstractmethod
    def residuals(self) -> None: ...

    @abstractmethod
    def predict_distribution(self, predictors: dict[str, float | int]) -> None:
        if not isinstance(predictors, dict):
            raise TypeError("Parameter 'predictors' must be a dictionary")

        if len(predictors) == 0:
            raise ValueError(
                "Parameter 'predictors' cannot be an empty dictionary"
            )

    @abstractmethod
    def likelihood(self, data: pd.DataFrame) -> None:
        if not isinstance(data, pd.DataFrame):
            raise TypeError(
                "Parameter 'data' must be an instance of 'pandas.DataFrame'"
            )

        if data.empty:
            raise ValueError(
                "Parameter 'data' cannot be an empty 'pandas.DataFrame'"
            )

    @abstractmethod
    def log_likelihood(self, data: pd.DataFrame) -> None:
        if not isinstance(data, pd.DataFrame):
            raise TypeError(
                "Parameter 'data' must be an instance of 'pandas.DataFrame'"
            )

        if data.empty:
            raise ValueError(
                "Parameter 'data' cannot be an empty 'pandas.DataFrame'"
            )

    @abstractmethod
    def _compute_model_parameters_at_posterior_means(self) -> None: ...

    @abstractmethod
    def _compute_model_parameters_at_observation(self, i: int) -> None: ...



from baypy.regression.functions import sample_sigma2
from baypy.regression.functions import sample_beta
from baypy.model import Model
from .regression import Regression
import numpy as np
from scipy.stats import norm, invgamma


class LinearRegression(Regression):
    r""":py:class:`LinearRegression <baypy.regression.linear_regression.LinearRegression>`
    object.

    Methods
    -------
    :meth:`sample`
        It samples a sequence of observations from the full posterior
        distribution of regressors' parameters :math:`\beta_j` and ``variance``
        :math:`\sigma^2`.

    .. admonition:: See Also
       :class: seealso

       :py:class:`LinearModel <baypy.model.linear_model.LinearModel>`
    """

    @staticmethod
    def sample(
        model: Model,
        n_iterations: int,
        burn_in_iterations: int,
        n_chains: int,
        seed: int = None
    ) -> None:
        r"""It samples a sequence of observations from the full posterior
        distribution of regressors' parameters :math:`\beta_j` and ``variance``
        :math:`\sigma^2`.
        First ``burn_in_iterations`` are discarded since they may not
        accurately represent the desired distribution.
        For each variable, it generates ``n_chain`` Markov chains.

        Parameters
        ----------
        ``model`` : :py:class:`Model <baypy.model.model.Model>`
            The model with data, regressors, response variable and priors to be
            solved through Monte Carlo sampling.
        ``n_iterations`` : :py:class:`int`
            Number of total sampling iteration for each chain. It must be a
            strictly positive integer.
        ``burn_in_iterations`` : :py:class:`int`
            Number of burn-in iteration for each chain. It must be a positive
            integer or ``0``.
        ``n_chains`` : :py:class:`int`
            Number of chains. It must be a strictly positive integer.
        ``seed`` : :py:class:`int`, optional
            Random seed to use for reproducibility of the sampling.

        .. admonition:: Raises
           :class: warning

           ``TypeError``
               - If ``model`` is not a
                 :py:class:`Model <baypy.model.model.Model>`,
               - if ``n_iterations`` is not an :py:class:`int`,
               - if ``burn_in_iterations`` is not an :py:class:`int`,
               - if ``n_chains`` is not an :py:class:`int`,
               - if ``seed`` is not an :py:class:`int`.
           ``ValueError``
               - If ``model.data`` is :py:obj:`None`,
               - if ``model.response_variable`` is :py:obj:`None`,
               - if ``model.response_variable`` is not a column of
                 ``model.data``
               - if ``model.priors`` is :py:obj:`None`,
               - if a ``model.priors`` key is not a column of ``model.data``,
               - If ``n_iterations`` is equal to or less than ``0``,
               - if ``burn_in_iterations`` is less than ``0``,
               - if ``n_chains`` is equal to or less than ``0``,
               - if ``seed`` is not between ``0`` and ``2**32 - 1``.

        .. admonition:: Notes
           :class: tip

           The linear regression model of the response variable :math:`y` with
           respect to regressors :math:`X` is:

           .. math::
               y \sim N(\mu, \sigma^2)
           .. math::
               \mu = \beta_0 + B X = \beta_0 + \sum_{j = 1}^m \beta_j x_j

           and the likelihood is:

           .. math::
               p \left( y \left\vert B,\sigma^2 \right. \right) =
               \frac{1}{\sqrt{2 \pi \sigma^2}} \exp{- \frac{\left(y -
               \mu \right)^2}{2 \sigma^2}} .
        """
        super(
            LinearRegression,
            LinearRegression
        ).sample(
            model=model,
            n_iterations=n_iterations,
            burn_in_iterations=burn_in_iterations,
            n_chains=n_chains,
            seed=seed
        )
        data = model.data.copy()

        regressor_names = model.variable_names.copy()
        regressor_names.pop(regressor_names.index('variance'))

        beta_0 = [model.priors[x]['mean'] for x in regressor_names]
        Beta_0 = np.array(beta_0)[np.newaxis].transpose()

        sigma_0 = [model.priors[x]['variance'] for x in regressor_names]
        Sigma_0 = np.zeros((len(sigma_0), len(sigma_0)))
        np.fill_diagonal(Sigma_0, sigma_0)
        Sigma_0_inv = np.linalg.inv(Sigma_0)

        k_0 = model.priors['variance']['shape']
        theta_0 = model.priors['variance']['scale']

        n = len(data)
        k_1 = k_0 + n

        y = data[model.response_variable]
        data['intercept'] = 1
        X = np.array(data[regressor_names])

        Xt_X = np.dot(X.transpose(), X)
        Xt_y = np.dot(X.transpose(), y)[np.newaxis].transpose()
        Sigma_0_inv_Beta_0 = np.dot(Sigma_0_inv, Beta_0)

        posteriors = {
            variable: [[] for _ in range(n_chains)]
            for variable in model.variable_names
        }

        if seed is not None:
            np.random.seed(seed)

        beta = [[norm.rvs(
            loc=model.priors[regressor]['mean'],
            scale=np.sqrt(model.priors[regressor]['variance'])
        ) for regressor in regressor_names] for _ in range(n_chains)]
        sigma2 = [invgamma.rvs(a=k_0, scale=theta_0) for _ in range(n_chains)]

        for _ in range(burn_in_iterations + n_iterations + 1):
            for k in range(n_chains):
                for j, regressor in enumerate(regressor_names, 0):
                    posteriors[regressor][k].append(beta[k][j])
                posteriors['variance'][k].append(sigma2[k])

            beta = [sample_beta(
                Xt_X=Xt_X,
                Xt_y=Xt_y,
                sigma2=sigma2[k],
                Sigma_0_inv=Sigma_0_inv,
                Sigma_0_inv_Beta_0=Sigma_0_inv_Beta_0
            ) for k in range(n_chains)]

            sigma2 = [sample_sigma2(
                y=y,
                X=X,
                beta=beta[k],
                k_1=k_1,
                theta_0=theta_0
            ) for k in range(n_chains)]

        model.posteriors = {
            posterior: np.array(
                posterior_samples
            ).transpose()[burn_in_iterations + 1:, :]
            for posterior, posterior_samples in posteriors.items()
        }



from ._functions import _compute_autocorrelation
import matplotlib.pyplot as plt
import numpy as np
from ..utils import flatten_matrix


def autocorrelation_plot(
    posteriors: dict[str, np.ndarray],
    max_lags: int = 30
) -> None:
    """It plots the auto-correlation for each Markov chain for each regression
    variable. \n
    The plot shows the auto-correlation trend from lag ``0`` (when
    auto-correlation is always ``1``) up to ``max_lags``.
    The plot layout has number of rows equal to the number of regression
    variables and a number of columns equal to the number of chains.

    Parameters
    ----------
    ``posteriors`` : :py:class:`dict`
        Posterior samples. Posteriors and relative samples are key-value pairs.
        Each sample is a :py:class:`numpy.ndarray` with a number of rows equal
        to the number of iterations and a number of columns equal to the number
        of Markov chains.
    ``max_lags`` : :py:class:`int`, optional
        Maximum number of lags to which compute the auto-correlation. The
        default is ``30``.

    .. admonition:: Raises
       :class: warning

       ``TypeError``
           - If ``posteriors`` is not a :py:class:`dict`,
           - if a posterior sample is not a :py:class:`numpy.ndarray`,
           - if ``max_lags`` is not an :py:class:`int`.
       ``KeyError``
           If ``posteriors`` does not contain ``'intercept'`` key.
       ``ValueError``
           - If a posterior sample is an empty :py:class:`numpy.ndarray`,
           - if ``max_lags`` is less or equal to ``0``.

    .. admonition:: See Also
       :class: seealso

       :py:func:`autocorrelation_summary <baypy.diagnostics.functions.autocorrelation_summary>` \n
       :py:func:`effective_sample_size <baypy.diagnostics.functions.effective_sample_size>`
    """
    if not isinstance(posteriors, dict):
        raise TypeError("Parameter 'posteriors' must be a dictionary")

    if not all([isinstance(posterior_sample, np.ndarray)
                for posterior_sample in posteriors.values()]):
        raise TypeError(
            "All posteriors data must be an instance of 'numpy.ndarray'"
        )

    if 'intercept' not in posteriors.keys():
        raise KeyError("Parameter 'posteriors' must contain a 'intercept' key")

    for posterior, posterior_samples in posteriors.items():
        if posterior_samples.size == 0:
            raise ValueError(f"Posterior '{posterior}' data is empty")

    if not isinstance(max_lags, int):
        raise TypeError("Parameter 'max_lags' must be a integer")
    if max_lags <= 0:
        raise ValueError("Parameter 'max_lags' must be greater than 0")

    variable_names = list(posteriors.keys())
    n_variables = len(variable_names)
    n_iterations, n_chains = posteriors['intercept'].shape

    _, ax = plt.subplots(
        nrows=n_variables,
        ncols=n_chains,
        figsize=(min(1.5*n_chains + 3, 10), min(1.5*n_variables + 2, 10)),
        sharex='all',
        sharey='all'
    )

    if n_chains > 1:
        for k in range(n_chains):
            ax[0, k].set_title(f"Chain {k + 1}")
            for j, variable in enumerate(variable_names, 0):
                acorr = _compute_autocorrelation(
                    vector=flatten_matrix(posteriors[variable][:, k]),
                    max_lags=max_lags
                )
                ax[j, k].stem(acorr, markerfmt=' ', basefmt=' ')
                ax[j, k].tick_params(
                    bottom=False,
                    top=False,
                    left=False,
                    right=False
                )

        for j, variable in enumerate(variable_names, 0):
            ax[j, 0].set_ylabel(variable)
            ax[j, 0].set_yticks([-1, 0, 1])

        ax[0, 0].set_xlim(-1, min(max_lags, n_iterations))
        ax[0, 0].set_ylim(-1, 1)

        plt.tight_layout()
        plt.subplots_adjust(left=0.1)

    else:
        ax[0].set_title('Chain 1')
        for j, variable in enumerate(variable_names, 0):
            acorr = _compute_autocorrelation(
                vector=flatten_matrix(posteriors[variable]),
                max_lags=max_lags
            )
            ax[j].stem(acorr, markerfmt=' ', basefmt=' ')

        for j, variable in enumerate(variable_names, 0):
            ax[j].set_ylabel(variable)
            ax[j].set_yticks([-1, 0, 1])

        ax[0].set_xlim(-1, min(max_lags, n_iterations))
        ax[0].set_ylim(-1, 1)

        plt.tight_layout()
        plt.subplots_adjust(left=0.14)

    plt.show()



__all__ = [
    "autocorrelation_plot",
    "autocorrelation_summary",
    "effective_sample_size"
]


from .autocorrelation_plot import autocorrelation_plot
from .autocorrelation_summary import autocorrelation_summary
from .effective_sample_size import effective_sample_size



__all__ = [
    "compute_DIC",
    "residuals_plot",
    "summary",
    "trace_plot"
]


from .compute_DIC import compute_DIC
from .residuals_plot import residuals_plot
from .summary import summary
from .trace_plot import trace_plot



from ._functions import _compute_autocorrelation
import numpy as np
import pandas as pd
from ..utils import flatten_matrix


def effective_sample_size(
    posteriors: dict[str, np.ndarray],
    print_summary: bool = True
) -> pd.DataFrame:
    """It computes and prints the effective number of sample for each
    posterior.

    Parameters
    ----------
    ``posteriors`` : :py:class:`dict`
        Posterior samples. Posteriors and relative samples are key-value pairs.
        Each sample is a :py:class:`numpy.ndarray` with a number of rows
        equals to the number of iterations and a number of columns equal to the
        number of Markov chains.
    ``print_summary`` : :py:class:`bool`, optional
        If ``True`` prints the effective sample size summary report. Default is
        ``True``.

    Returns
    -------
    :py:class:`pandas.DataFrame`
        The dataframe with a single row and a number of columns equal to the
        number of model variables. The unique index of the dataframe is
        ``'Effective Sample Size'``.

    .. admonition:: Raises
       :class: warning

       ``TypeError``
           - If ``posteriors`` is not a :py:class:`dict`,
           - if a posterior sample is not a :py:class:`numpy.ndarray`,
           - if ``print_summary`` is not a :py:class:`bool`.
       ``KeyError``
           If ``posteriors`` does not contain ``'intercept'`` key.
       ``ValueError``
           If a posterior sample is an empty :py:class:`numpy.ndarray`.

    .. admonition:: Notes
       :class: tip

       The effective number of sample could be theoretically equal to the
       number of iterations in case of no auto-correlation of the Markov chain.
       The greater the auto-correlation of the Markov chain, the smaller the
       effective sample size of the posterior.

    .. admonition:: See Also
       :class: seealso

       :py:func:`autocorrelation_plot <baypy.diagnostics.functions.autocorrelation_plot>` \n
       :py:func:`autocorrelation_summary <baypy.diagnostics.functions.autocorrelation_summary>`
    """
    if not isinstance(posteriors, dict):
        raise TypeError("Parameter 'posteriors' must be a dictionary")

    if not all(
        [
            isinstance(posterior_sample, np.ndarray)
            for posterior_sample in posteriors.values()
        ]
    ):
        raise TypeError(
            "All posteriors data must be an instance of 'numpy.ndarray'"
        )

    if not isinstance(print_summary, bool):
        raise TypeError("Parameter 'print_summary' must be a boolean")

    if 'intercept' not in posteriors.keys():
        raise KeyError("Parameter 'posteriors' must contain a 'intercept' key")

    for posterior, posterior_samples in posteriors.items():
        if posterior_samples.size == 0:
            raise ValueError(f"Posterior '{posterior}' data is empty")

    n_chains = posteriors['intercept'].shape[1]
    ess_summary = pd.DataFrame(
        columns=list(posteriors.keys()),
        index=["Effective Sample Size"]
    )

    for variable in ess_summary.columns:
        variable_ess = []
        for k in range(n_chains):
            vector = flatten_matrix(posteriors[variable][:, k])
            n = len(vector)
            variable_chain_acorr = _compute_autocorrelation(
                vector=vector,
                max_lags=n
            )
            indexes = np.arange(2, len(variable_chain_acorr), 1)
            indexes = indexes[
                (variable_chain_acorr[1:-1] + variable_chain_acorr[2:] < 0) &
                (indexes % 2 == 1)
            ]
            index = indexes[0] if indexes.size > 0 \
                else len(variable_chain_acorr) + 1
            ess = n/(1 + 2*np.abs(variable_chain_acorr[1:index - 1].sum()))
            variable_ess.append(ess)

        ess_summary[variable] = np.sum(variable_ess)

    if print_summary:
        with pd.option_context('display.precision', 2):
            print(ess_summary.to_string())

    return ess_summary



import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import gaussian_kde


def trace_plot(posteriors: dict[str, np.ndarray]) -> None:
    """It plots the traces and the probability density for each posterior. \n
    The plot shows the traces for each Markov chain, for each regression
    variable and the relative posterior density.
    The plot layout has number of rows equal to the number of regression
    variables and two columns: traces on the left and densities on the right.

    Parameters
    ----------
    ``posteriors`` : :py:class:`dict`
        Posterior samples. Posteriors and relative samples are key-value
        pairs. Each sample is a :py:class:`numpy.ndarray` with a number of rows
        equal to the number of iterations and a number of columns equal to the
        number of Markov chains.

    .. admonition:: Raises
       :class: warning

       ``TypeError``
           - If ``posteriors`` is not a :py:class:`dict`,
           - if a posterior sample is not a :py:class:`numpy.ndarray`.
       ``KeyError``
           If ``posteriors`` does not contain ``'intercept'`` key.
       ``ValueError``
           If a posterior sample is an empty :py:class:`numpy.ndarray`.

    .. admonition:: See Also
       :class: seealso

       :py:class:`LinearRegression <baypy.regression.linear_regression.LinearRegression>`
    """
    if not isinstance(posteriors, dict):
        raise TypeError("Parameter 'posteriors' must be a dictionary")

    if not all(
        [
            isinstance(posterior_sample, np.ndarray)
            for posterior_sample in posteriors.values()
        ]
    ):
        raise TypeError(
            "All posteriors data must be an instance of 'numpy.ndarray'"
        )

    if 'intercept' not in posteriors.keys():
        raise KeyError("Parameter 'posteriors' must contain a 'intercept' key")

    for posterior, posterior_samples in posteriors.items():
        if posterior_samples.size == 0:
            raise ValueError(f"Posterior '{posterior}' data is empty")

    variable_names = list(posteriors.keys())
    n_variables = len(variable_names)
    n_iterations = len(posteriors['intercept'])

    fig = plt.figure(figsize=(10, min(1.5*n_variables + 2, 10)))
    trace_axes = []
    for j, variable in zip(range(1, 2*n_variables, 2), variable_names):
        ax_j_trace = fig.add_subplot(n_variables, 2, j)
        ax_j_density = fig.add_subplot(n_variables, 2, j + 1)

        ax_j_trace.plot(posteriors[variable], linewidth=0.5)
        ax_j_density.plot(*_compute_kde(posteriors[variable].flatten()))

        ax_j_trace.set_title(f"Trace of {variable} parameter")
        ax_j_density.set_title(f"Density of {variable} parameter")
        ax_j_trace.tick_params(
            bottom=False,
            top=False,
            left=False,
            right=False
        )
        ax_j_density.tick_params(
            bottom=False,
            top=False,
            left=False,
            right=False
        )
        ax_j_density.set_ylim(0, )
        trace_axes.append(ax_j_trace)

    for ax_j in trace_axes[1:]:
        ax_j.sharex(trace_axes[0])
    trace_axes[0].set_xlim(0, n_iterations - 1)

    plt.tight_layout()

    plt.show()


def _compute_kde(posterior: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    posterior_support = np.linspace(np.min(posterior), np.max(posterior), 1000)
    posterior_kde = gaussian_kde(posterior)(posterior_support)

    return posterior_support, posterior_kde



from ._functions import _compute_autocorrelation
import numpy as np
import pandas as pd
from typing import Iterable
from ..utils import flatten_matrix


def autocorrelation_summary(
    posteriors: dict[str, np.ndarray],
    lags: Iterable[int] = None,
    print_summary: bool = True
) -> pd.DataFrame:
    """It prints the auto-correlation summary for each regression variable. \n
    The summary reports the auto-correlation values at the lags listed in
    ``lags``.

    Parameters
    ----------
    ``posteriors`` : :py:class:`dict`
        Posterior samples. Posteriors and relative samples are key-value pairs.
        Each sample is a :py:class:`numpy.ndarray` with a number of rows equal
        to the number of iterations and a number of columns equal to the number
        of Markov chains.
    ``lags`` : :py:class:`Iterable <typing.Iterable>`, optional
        List of the lags to which compute the auto-correlation. It cannot be an
        empty :py:class:`Iterable <typing.Iterable>`. It must contain only
        positive integers. The default is ``[0, 1, 5, 10, 30]``.
    ``print_summary`` : :py:class:`bool`, optional
        If ``True`` prints the autocorrelation summary report. Default is
        ``True``.

    Returns
    -------
    :py:class:`pandas.DataFrame`
        The dataframe with a number of row equal to the number of element in
        ``lags`` and a number of columns equal to the number of model
        variables. Lags are reported in dataframe index.

    .. admonition:: Raises
       :class: warning

       ``TypeError``
           - If ``posteriors`` is not a :py:class:`dict`,
           - if a posterior sample is not a :py:class:`numpy.ndarray`,
           - if ``lags`` is not an :py:class:`Iterable <typing.Iterable>`,
           - if ``lags`` does not contain only :py:class:`int`,
           - if ``print_summary`` is not a :py:class:`bool`.
       ``KeyError``
           If ``posteriors`` does not contain ``'intercept'`` key.
       ``ValueError``
           - If a posterior sample is an empty :py:class:`numpy.ndarray`,
           - if ``lags`` is an empty :py:class:`Iterable <typing.Iterable>`,
           - if a value in ``lags`` is a negative :py:class:`int`.

    .. admonition:: Notes
       :class: tip

       The reported auto-correlation for each variable is a mean of
       auto-correlations for the chains of that variable, for each chain.

    .. admonition:: See Also
       :class: seealso

       :py:func:`autocorrelation_plot <baypy.diagnostics.functions.autocorrelation_plot>` \n
       :py:func:`effective_sample_size <baypy.diagnostics.functions.effective_sample_size>`
    """
    if not isinstance(posteriors, dict):
        raise TypeError("Parameter 'posteriors' must be a dictionary")

    if not all(
        [
            isinstance(posterior_sample, np.ndarray)
            for posterior_sample in posteriors.values()
        ]
    ):
        raise TypeError(
            "All posteriors data must be an instance of 'numpy.ndarray'"
        )

    if not isinstance(print_summary, bool):
        raise TypeError("Parameter 'print_summary' must be a boolean")

    if 'intercept' not in posteriors.keys():
        raise KeyError("Parameter 'posteriors' must contain a 'intercept' key")

    for posterior, posterior_samples in posteriors.items():
        if posterior_samples.size == 0:
            raise ValueError(f"Posterior '{posterior}' data is empty")

    if lags is not None:
        if not isinstance(lags, Iterable):
            raise TypeError("Parameter 'lags' must be an iterable")
        if not lags:
            raise ValueError("Parameter 'lags' cannot be an empty iterable")
        if not all([isinstance(lag, int) for lag in lags]):
            raise TypeError("Parameter 'lags' must contain only integers")
        if any([lag < 0 for lag in lags]):
            raise ValueError(
                "Parameter 'lags' cannot contain negative integers"
            )

    lags = [0, 1, 5, 10, 30] if lags is None else list(set(lags))

    n_chains = posteriors['intercept'].shape[1]
    acorr_summary = pd.DataFrame(
        columns=list(posteriors.keys()),
        index=[f"Lag {lag}" for lag in lags]
    )

    for variable in acorr_summary.columns:
        variable_acorr = []
        for k in range(n_chains):
            variable_chain_acorr = _compute_autocorrelation(
                vector=flatten_matrix(posteriors[variable][:, k]),
                max_lags=max(lags) + 1
            )
            variable_acorr.append(variable_chain_acorr[lags])
        variable_acorr = np.array(variable_acorr)
        acorr_summary[variable] = variable_acorr.mean(axis=0)

    if print_summary:
        print(acorr_summary.to_string())

    return acorr_summary



import numpy as np


def _compute_autocorrelation(vector: np.ndarray, max_lags: int) -> np.ndarray:

    normalized_vector = vector - vector.mean()
    autocorrelation = np.correlate(
        normalized_vector,
        normalized_vector, 'full'
    )[len(normalized_vector) - 1:]
    autocorrelation = autocorrelation/vector.var()/len(vector)
    autocorrelation = autocorrelation[:max_lags]

    return autocorrelation
